#include <unistd.h>

#include <fstream>
#include <future>
#include <iostream>
#include <memory>
#include <string>
#include <thread>
#include <vector>

#include "include/slam_manager.h"
#include "include/data/package_line.h"
#include "int_test.h"
#include "SLAM_GeneralWrapper.h"
#include "slam/executable.h"

void package_ins_inputs(slam::InsData *const _insData, CAN2SWC const _canData);
void package_rm_outputs(dpc::slam_road_model *const out_rmData, slam::slam_road_model const in_rmData);

int main(int argc, char **argv)
{
    // setup_test();

    // Logger for wrapper
    // ara headers must be added

    ara::log::InitLogging("DPC", "SLAM_TEST", ara::log::LogLevel::kVerbose, (ara::log::LogMode::kConsole | ara::log::LogMode::kRemote));
    ara::log::Logger &mainLog{ara::log::CreateLogger("main", "slam sample context", ara::log::LogLevel::kVerbose)};
    mainLog.LogInfo() << "Let's produce some slam data...";

    GeneralWrapper gw(mainLog);

    // output
    //***to road model
    dpc::slam_road_model slam_road_model_data;

    //***to decision
    dpc::ControlWaypoints slam_con_wayp_data;

    dpc::slam_odometry slam_odom_data;

    // input:
    CAN2SWC can2slam_data; // Executable Sub data

    // Local variables and object declarations
    uint32_t slam_run_counter = 0;
    slam::PackageLine wk_package_line;

    slam::slam_road_model wk_rm_;
    slam::slam_odometry wk_odom;
    slam::ControlWaypoints wk_cont_way_points;

    wk_rm_.lanes.resize(1);
    wk_rm_.lines.resize(1);
    wk_rm_.lines[0].points.resize(100);

    while (1)
    {
        if (slam_run_counter == 0)
        {

            wk_package_line.Init();
            printf("slam initialized!\n");

            slam_road_model_data.cycle_counter = 0;
            slam_con_wayp_data.msg_header.cycle_counter = 0;
        }
        else
        {
            //********************************Get input data********************************
            gw._CAN2SLAM_Required.s.GetData(can2slam_data);
            //           std::cout << "can data: " << can2slam_data.VCU_Info_1_Signals.ses_en_sts << std::endl;

            //            ********************************run slam program********************************
            slam::InsData _insdata;
            package_ins_inputs(&_insdata, can2slam_data);
            wk_package_line.SetHostVehiclePosition(_insdata);
            bool wk_odom_flag = wk_package_line.PackageOdom(wk_odom);
            bool wk_rm_flag = wk_package_line.PackageRM(wk_rm_.lines[0].points);
            wk_rm_.lanes[0].lane_id = 1;
            wk_rm_.lanes[0].lane_no = 2;
            wk_rm_.lanes[0].center_line_id = 10;
            // wk_rm_.lanes[0].speed_limit_value = 50;

            wk_rm_.lines[0].id = 10;
            //********************************publish output data********************************
            slam_road_model_data.cycle_counter++;
            package_rm_outputs(&slam_road_model_data, wk_rm_);
            gw._SLAM2EMslam_road_model_Provide.s.SetData(slam_road_model_data);
            gw._SLAM2EMslam_road_model_Provide.Publish();

            //SLAM2DMControlWaypoints_Provide _SLAM2DMControlWaypoints_Provide;
            slam_con_wayp_data.msg_header.cycle_counter++;
            gw._SLAM2DMControlWaypoints_Provide.s.SetData(slam_con_wayp_data);
            gw._SLAM2DMControlWaypoints_Provide.Publish();

            // Slam2PerceptionOdom_Provide _Slam2PerceptionOdom_Provide;
            slam_odom_data.cycle_counter++;
            slam_odom_data.utc_timestamp = wk_odom.utc_timestamp;
            slam_odom_data.position.position.x = wk_odom.position.position.x;
            slam_odom_data.position.position.y = wk_odom.position.position.y;
            slam_odom_data.position.position.z = wk_odom.position.position.z;

            slam_odom_data.position.orientation.x = wk_odom.position.orientation.x;
            slam_odom_data.position.orientation.y = wk_odom.position.orientation.y;
            slam_odom_data.position.orientation.z = wk_odom.position.orientation.z;
            slam_odom_data.position.orientation.w = wk_odom.position.orientation.w;

            gw._Slam2PerceptionOdom_Provide.s.SetData(slam_odom_data);
            gw._Slam2PerceptionOdom_Provide.Publish();
        }

        slam_run_counter++;

        std::this_thread::sleep_for(std::chrono::milliseconds(20)); // slam cycle time: 20 ms
    }

    return 0;
}

void package_ins_inputs(slam::InsData *const _insData, CAN2SWC const _canData)
{
    _insData->utc_timestamp = _canData.INS_SOL_RESULT_LONGITUDE_Signals.timestamp_s * 1e3 + _canData.INS_SOL_RESULT_LONGITUDE_Signals.timestamp_ns * 1e-6; // milliseconds

    _insData->longitude = _canData.INS_SOL_RESULT_LONGITUDE_Signals.ins_longitude;
    _insData->latitude = _canData.INS_SOL_RESULT_LATITUDE_Signals.ins_latitude;
    _insData->heading = _canData.INS_SOL_RESULT_ANGLE_Signals.ins_yaw_angle;

    //_insData->status
}

void package_rm_outputs(dpc::slam_road_model *const out_rmData,
                        slam::slam_road_model const in_rmData)
{

    out_rmData->lanes.resize(in_rmData.lanes.size());
    for (int i = 0; i < in_rmData.lanes.size(); ++i)
    {
        dpc::traffic_lane _lane;
        _lane.lane_id = in_rmData.lanes[i].lane_id;
        _lane.center_line_id = in_rmData.lanes[i].center_line_id;
        _lane.lane_no = in_rmData.lanes[i].lane_no;
        // _lane.speed_limit_value = in_rmData.lanes[i].speed_limit_value;
        out_rmData->lanes[i] = _lane;
    }


    out_rmData->lines.resize(in_rmData.lines.size());

    for (int i = 0; i < in_rmData.lines.size(); ++i)
    {
        out_rmData->lines[i].points.resize(in_rmData.lines[i].points.size());

        out_rmData->lines[i].id = in_rmData.lines[i].id;

        for (int j = 0; j < in_rmData.lines[i].points.size(); ++j)
        {
            dpc::line_point _p;
            _p.point.x = in_rmData.lines[i].points[j].point.x;
            _p.point.y = in_rmData.lines[i].points[j].point.y;

            out_rmData->lines[i].points[j] = _p;
        }
    }
}
